var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
import mapSeries from "p-each-series";
import PQueue from "p-queue";
import { Log, JSON_ENCODING, } from "@dao-xyz/peerbit-log";
import { Entry } from "@dao-xyz/peerbit-log";
import { stringifyCid } from "@dao-xyz/peerbit-block";
import { variant, option, field, vec } from "@dao-xyz/borsh";
import { serialize, deserialize } from "@dao-xyz/borsh";
import { Snapshot } from "./snapshot.js";
import { AccessError, } from "@dao-xyz/peerbit-crypto";
import { waitForAsync } from "@dao-xyz/peerbit-time";
import { logger as loggerFn } from "@dao-xyz/peerbit-logger";
import path from "path-browserify";
import { v4 as uuid } from "uuid";
import { join } from "./replicator.js";
const logger = loggerFn({ module: "store" });
export class CachedValue {
}
let CID = class CID extends CachedValue {
    hash;
    constructor(opts) {
        super();
        if (opts) {
            this.hash = opts.hash;
        }
    }
};
__decorate([
    field({ type: "string" }),
    __metadata("design:type", String)
], CID.prototype, "hash", void 0);
CID = __decorate([
    variant(0),
    __metadata("design:paramtypes", [Object])
], CID);
export { CID };
let UnsfinishedReplication = class UnsfinishedReplication extends CachedValue {
    hashes;
    constructor(opts) {
        super();
        if (opts) {
            this.hashes = opts.hashes;
        }
    }
};
__decorate([
    field({ type: vec("string") }),
    __metadata("design:type", Array)
], UnsfinishedReplication.prototype, "hashes", void 0);
UnsfinishedReplication = __decorate([
    variant(1),
    __metadata("design:paramtypes", [Object])
], UnsfinishedReplication);
export { UnsfinishedReplication };
let HeadsCache = class HeadsCache extends CachedValue {
    heads;
    last;
    constructor(heads, last) {
        super();
        this.heads = heads;
        this.last = last;
    }
};
__decorate([
    field({ type: vec("string") }),
    __metadata("design:type", Array)
], HeadsCache.prototype, "heads", void 0);
__decorate([
    field({ type: option("string") }),
    __metadata("design:type", String)
], HeadsCache.prototype, "last", void 0);
HeadsCache = __decorate([
    variant(2),
    __metadata("design:paramtypes", [Array, String])
], HeadsCache);
export { HeadsCache };
export const DefaultOptions = {
    maxHistory: -1,
    replicationConcurrency: 32,
    typeMap: {},
};
let Store = class Store {
    _storeIndex; // how to ensure unqiueness
    _canAppend;
    _onUpdate;
    _onUpdateOption;
    // An access controller that is note part of the store manifest, usefull for circular store -> access controller -> store structures
    _options;
    identity;
    /*   events: EventEmitter;
     */
    headsPath;
    snapshotPath;
    initialized;
    encoding = JSON_ENCODING;
    _store;
    _cache;
    _oplog;
    _queue;
    _stats;
    _key;
    _saveFile;
    _loadFile;
    constructor(properties) {
        if (properties) {
            this._storeIndex = properties?.storeIndex;
        }
    }
    setup(properties) {
        this.encoding = properties.encoding;
        this.onUpdate = properties.onUpdate;
        this.canAppend = properties.canAppend;
    }
    async init(store, identity, options) {
        if (this.initialized) {
            throw new Error("Already initialized");
        }
        this._saveFile =
            options.saveFile || ((file) => store.put(file, "dag-cbor"));
        this._loadFile = options.loadFile || ((file) => store.get(file));
        // Set ipfs since we are to save the store
        this._store = store;
        // Set the options (we will use the topic property after thiis)
        const opts = { ...DefaultOptions, ...options };
        this._options = opts;
        // Create IDs, names and paths
        this.identity = identity;
        this._onUpdateOption = options.onUpdate;
        /* this.events = new EventEmitter() */
        this.headsPath = path.join(this.id, "_heads");
        this.snapshotPath = path.join(this.id, "snapshot");
        // External dependencies
        this._cache = await this._options.resolveCache(this);
        // Create the operations log
        this._oplog = new Log(this._store, identity, this.logOptions);
        // _addOperation and log-joins queue. Adding ops and joins to the queue
        // makes sure they get processed sequentially to avoid race conditions
        // between writes and joins (coming from Replicator)
        this._queue = new PQueue({ concurrency: 1 });
        this._stats = {
            snapshot: {
                bytesLoaded: -1,
            },
            syncRequestsReceieved: 0,
        };
        if (this._options.onOpen) {
            await this._options.onOpen(this);
        }
        this.initialized = true;
        return this;
    }
    async updateHeadsCache(newHeads, reset) {
        // If 'reset' then dont keep references to old heads caches, assume new cache will fully describe all heads
        const last = reset
            ? undefined
            : await this._cache.get(this.headsPath);
        const newHeadsPath = path.join(this.headsPath, uuid());
        await this._cache.set(this.headsPath, newHeadsPath);
        await this._cache.setBinary(newHeadsPath, new HeadsCache(newHeads, last));
    }
    get id() {
        if (typeof this._storeIndex !== "number") {
            throw new Error("Store index not set");
        }
        return this._storeIndex.toString();
    }
    get oplog() {
        return this._oplog;
    }
    get key() {
        return this._key;
    }
    get logOptions() {
        return {
            logId: this.id,
            encryption: this._options.encryption,
            encoding: this.encoding,
            sortFn: this._options.sortFn,
            prune: this._options.prune,
        };
    }
    setIdentity(identity) {
        this.identity = identity;
        this._oplog.setIdentity(identity);
    }
    set canAppend(canAppend) {
        this._canAppend = canAppend;
    }
    get canAppend() {
        return this._canAppend;
    }
    set onUpdate(onUpdate) {
        this._onUpdate = onUpdate;
    }
    async close() {
        if (!this.initialized) {
            return;
        }
        // Wait for the operations queue to finish processing
        // to make sure everything that all operations that have
        // been queued will be written to disk
        await this._queue?.onIdle();
        // Reset database statistics
        this._stats = {
            snapshot: {
                bytesLoaded: -1,
            },
            syncRequestsReceieved: 0,
        };
        if (this._options.onClose) {
            await this._options.onClose(this);
        }
        this._oplog = null;
        // Database is now closed
        return Promise.resolve();
    }
    /**
     * Drops a database and removes local data
     */
    async drop() {
        if (!this._oplog && !this._cache) {
            return; // already dropped
        }
        if (this._options.onDrop) {
            await this._options.onDrop(this);
        }
        await this._cache.del(this.headsPath);
        await this._cache.del(this.snapshotPath);
        await this.close();
        // Reset
        // TODO fix types
        this._oplog = undefined;
        this._cache = undefined;
        this.initialized = false; // call this last because (close() expect initialized to be able to function)
    }
    async load(amount, opts = {}) {
        if (!this.initialized) {
            throw new Error("Store needs to be initialized before loaded");
        }
        amount = amount || this._options.maxHistory;
        const fetchEntryTimeout = opts.fetchEntryTimeout || this._options.fetchEntryTimeout;
        if (this._options.onLoad) {
            await this._options.onLoad(this);
        }
        const heads = await this.getCachedHeads();
        // Load the log
        const log = await Log.fromEntryHash(this._store, this.identity, heads, {
            ...this.logOptions,
            length: amount,
            timeout: fetchEntryTimeout,
            onFetched: this._onLoadProgress.bind(this),
            concurrency: this._options.replicationConcurrency,
        });
        this._oplog = log;
        // Update the index
        if (heads.length > 0) {
            await this._updateIndex();
        }
        this._options.onReady && this._options.onReady(this);
    }
    /**
     *
     * @param entries
     * @returns change
     */
    async sync(entries, options = { save: true }) {
        this._stats.syncRequestsReceieved += 1;
        logger.debug(`Sync request #${this._stats.syncRequestsReceieved} ${entries.length}`);
        if (entries.length === 0) {
            return false;
        }
        const handle = async (headToHandle) => {
            const allEntries = headToHandle instanceof Entry
                ? [headToHandle]
                : [headToHandle.entry, ...headToHandle.references];
            await Promise.all(allEntries.map((h) => h.init({
                encoding: this.oplog._encoding,
                encryption: this.oplog._encryption,
            })));
            const entry = headToHandle instanceof Entry
                ? headToHandle
                : headToHandle.entry;
            this._options.onReplicationQueued &&
                this._options.onReplicationQueued(this, entry);
            const canAppend = options?.canAppend || this.canAppend;
            if (canAppend && !(await canAppend(entry))) {
                logger.debug("Not allowd to append head " + entry.hash);
                return Promise.resolve(null);
            }
            await Promise.all(allEntries.map(async (head) => {
                const headHash = head.hash;
                head.hash = undefined;
                try {
                    const hash = options?.save
                        ? await this._store.put(serialize(head), "raw")
                        : stringifyCid((await this._store.block(serialize(head), "raw")).cid);
                    head.hash = headHash;
                    if (head.hash === undefined) {
                        head.hash = hash; // can happen if you sync entries that you load directly from ipfs
                    }
                    else if (hash !== head.hash) {
                        logger.error("Head hash didn't match the contents");
                        throw new Error("Head hash didn't match the contents");
                    }
                }
                catch (error) {
                    logger.error(error);
                    throw error;
                }
            }));
            return headToHandle;
        };
        const hash = (entry) => {
            if (entry instanceof Entry) {
                return entry.hash;
            }
            else if (typeof entry === "string") {
                return entry;
            }
            return entry.entry.hash;
        };
        const newEntries = [];
        for (const entry of entries) {
            const h = hash(entry);
            if (h && this.oplog.has(h)) {
                continue;
            }
            newEntries.push(typeof entry === "string"
                ? await Entry.fromMultihash(this._store, entry)
                : entry);
        }
        if (newEntries.length === 0) {
            return false;
        }
        const saved = await mapSeries(newEntries, handle);
        const { change } = await join(saved, this._oplog, {
            concurrency: this._options.replicationConcurrency,
            onFetched: (entry) => this._options.onReplicationFetch &&
                this._options.onReplicationFetch(this, entry),
        });
        // TODO add head cache 'reset' so that it becomes more accurate over time
        await this.updateHeadsCache(newEntries.map((x) => hash(x)));
        await this._updateIndex(change);
        this._options.onReplicationComplete &&
            this._options.onReplicationComplete(this);
        return true;
    }
    get replicate() {
        return !!this._options.replicate;
    }
    async getCachedHeads() {
        if (!this._cache) {
            return [];
        }
        const result = [];
        let next = await this._cache.get(this.headsPath);
        while (next) {
            const cache = await this._cache.getBinary(next, HeadsCache);
            next = cache?.last;
            cache?.heads.forEach((head) => {
                result.push(head);
            });
        }
        return result;
    }
    async saveSnapshot() {
        const snapshotData = this._oplog.toSnapshot();
        const buf = serialize(new Snapshot({
            id: snapshotData.id,
            heads: snapshotData.heads,
            size: BigInt(snapshotData.values.length),
            values: snapshotData.values,
        }));
        const snapshot = await this._saveFile(buf);
        await this._cache.setBinary(this.snapshotPath, new CID({ hash: snapshot }));
        await waitForAsync(async () => (await this._cache.getBinary(this.snapshotPath, CID)) !==
            undefined, { delayInterval: 200, timeout: 10 * 1000 });
        logger.debug(`Saved snapshot: ${snapshot}`);
        return [snapshot];
    }
    async loadFromSnapshot() {
        if (this._options.onLoad) {
            await this._options.onLoad(this);
        }
        await this.sync([]);
        const snapshotCID = await this._cache.getBinary(this.snapshotPath, CID);
        if (snapshotCID) {
            const file = await this._loadFile(snapshotCID.hash);
            if (!file) {
                throw new Error("Missing snapshot");
            }
            const snapshotData = deserialize(file, Snapshot);
            // Fetch the entries
            // Timeout 1 sec to only load entries that are already fetched (in order to not get stuck at loading)
            if (snapshotData) {
                this._oplog = await Log.fromEntry(this._store, this.identity, snapshotData.heads, {
                    sortFn: this._options.sortFn,
                    length: -1,
                    timeout: 1000,
                    onFetched: this._onLoadProgress.bind(this),
                });
                await this._updateIndex();
                this._options.onReplicationComplete &&
                    this._options.onReplicationComplete(this);
            }
            this._options.onReady && this._options.onReady(this);
        }
        else {
            throw new Error(`Snapshot for ${this.id} not found!`);
        }
        return this;
    }
    async _updateIndex(entries) {
        // TODO add better error handling
        try {
            if (this._onUpdate) {
                await this._onUpdate(this._oplog, entries);
            }
        }
        catch (error) {
            if (error instanceof AccessError) {
                // fail silently for now
                logger.info("Could not update index due to AccessError");
            }
            else {
                throw error;
            }
        }
        try {
            if (this._onUpdateOption) {
                await this._onUpdateOption(this._oplog, entries);
            }
        }
        catch (error) {
            if (error instanceof AccessError) {
                // fail silently for now
                logger.info("Could not update index due to AccessError");
            }
            else {
                throw error;
            }
        }
    }
    async _addOperation(data, options) {
        const isReferencingAllHeads = !options?.nexts;
        const entry = await this._oplog.append(data, {
            nexts: options?.nexts,
            pin: options?.pin,
            reciever: options?.reciever,
            canAppend: options?.skipCanAppendCheck ? undefined : this.canAppend,
            identity: options?.identity,
            prune: options?.prune,
        });
        logger.debug("Appended entry with hash: " + entry.hash);
        await this.updateHeadsCache([entry.hash], isReferencingAllHeads);
        await this._updateIndex([entry]);
        // The row below will emit an "event", which is subscribed to on the orbit-db client (confusing enough)
        // there, the write is binded to the pubsub publish, with the entry. Which will send this entry
        // to all the connected peers to tell them that a new entry has been added
        // TODO: don't use events, or make it more transparent that there is a vital subscription in the background
        // that is handling replication
        this._options.onWrite && this._options.onWrite(this, entry);
        return entry;
    }
    /* Loading progress callback */
    _onLoadProgress(entry) {
        this._options.onLoadProgress &&
            this._options.onLoadProgress(this, entry);
    }
    clone() {
        return deserialize(serialize(this), this.constructor);
    }
};
__decorate([
    field({ type: "u32" }),
    __metadata("design:type", Number)
], Store.prototype, "_storeIndex", void 0);
Store = __decorate([
    variant(0),
    __metadata("design:paramtypes", [Object])
], Store);
export { Store };
//# sourceMappingURL=store.js.map